import click
import numpy as np
import rasterio
import warnings

from .worker import Worker
from pysharpen.methods import BroveyPansharpening, IHSPansharpening, GIHSPansharpening, LinearBrightnessScale
from pysharpen.functional import saturate_cast

METHODS= {'ihs' : IHSPansharpening,
          'gihs': GIHSPansharpening,
          'brovey': BroveyPansharpening}
PREPROCESSINGS = ['minmax', 'meanstd', 'none']


def process_nogeo(pan_file, ms_file, out_file,
                   methods, out_dtype, resampling='bilinear'):
    """
    A function to process the images without georeference (or without regard to existing georeference)
    The multispectral file is simply resampled to the panchrom image size and read as whole.

    Args:
        pan_file:
        ms_file:
        out_file:
        methods:
        out_dtype:
        resampling:

    Returns:

    """
    with rasterio.open(pan_file) as pan:
        profile = pan.profile
        pan_img = pan.read(1)
    dtype = out_dtype or profile['dtype']
    with rasterio.open(ms_file) as ms:
        ms_img = np.zeros((ms.count, pan_img.shape[0], pan_img.shape[1]), dtype=ms.profile['dtype'])
        ms.read(out=ms_img, resampling=getattr(rasterio.enums.Resampling, resampling))
        profile.update(count=ms.count, dtype=dtype)
        colorinterp = ms.colorinterp

    for method in methods:
        method.setup_from_patch(pan_img, ms_img)
        method.finalize_setup()
    for method in methods:
        pan_img, ms_img = method.process(pan_img, ms_img)

    with rasterio.open(out_file, 'w', **profile) as dst:
        dst.colorinterp = colorinterp
        dst.write(saturate_cast(ms_img, dtype))


@click.command()
@click.argument('pan_path', type=click.Path(exists=True))
@click.argument('ms_path', type=click.Path(exists=True))
@click.argument('out_path', type=click.Path(exists=False, writable=True))
@click.option('--method', type=str, required=False, default='ihs')
@click.option('--resampling', type=str, required=False, default='bilinear')
@click.option('--preprocessing', type=click.Choice(PREPROCESSINGS), required=False, default='none')
@click.option('--nogeo', is_flag=True, help='Ignores georeference of the image, and resamples MS to PAN size')
@click.option('--noclean', is_flag=True, help='Leave the intermediate files containing separate spectral bands')
@click.option("-v", is_flag=True, help='Enables verbose output')
#@click.option("--separate", is_flag=True, help='Used the channels are in separate files')
def command(pan_path, ms_path, out_path,
            method='ihs',
            resampling='bilinear',
            preprocessing='none',
            nogeo=False, noclean=False, v=False):
    """
    CLI: pysharpen panchrom_name.tif multispectral_name.tif out_name.tif method <preprocessing>
    method = <ihs|brovey>
    """
    if not v:
        warnings.filterwarnings("ignore")
    methods = []
    out_dtype=None
    # We provide an example of the most basic preprocessing functionality: panchrom is scaled separately,
    # all the spectral channels together
    if preprocessing == 'minmax':
        out_dtype='uint8'
        methods.append(LinearBrightnessScale('minmax', dtype='uint8', pan_separate=True, per_channel=False))
    elif preprocessing == 'meanstd':
        out_dtype='uint8'
        methods.append(LinearBrightnessScale('meanstd', dtype='uint8', pan_separate=True, per_channel=False))

    methods.append(METHODS[method]())

    if nogeo:
        if v:
            print('Start processing of not georeferenced images')
        process_nogeo(pan_path, ms_path, out_path,
                      methods=methods, resampling=resampling, out_dtype=out_dtype)

    else:
        if v:
            print('Start tile-based processing of georeferenced images')
        # todo: detect broken georeference and use nogeo option in this case?
        w = Worker(methods=methods, resampling=resampling, out_dtype=out_dtype)
        w.process_single(pan_path, ms_path, out_path, clean=not noclean, verbose=v)
    if v:
        print('Finished')

if __name__ == "__main__":
    command()


import numpy as np
from math import sqrt
from pysharpen.methods.img_proc import ImgProc
from pysharpen.functional import linear_brightness_scale

ALLOWED_METHODS = ['minmax', 'meanstd', 'histogram']


class LinearBrightnessScale(ImgProc):
    """
     IHS pansharpening is for RGB image only. It transforms RGB into IHS color space,
     then replaces Intensity with PAN channel and transforms backwards. Based on OpenCV color transforms
    """

    def __init__(self, method='minmax', dtype=None,
                 per_channel=False, pan_separate=False,
                 process_pan=True, process_ms=True,
                 std_width=3.0, hist_cut=0.05):
        """

        Args:
            method: minmax
            dtype:
            per_channel:
            pan_separate:
            process_pan: can be disabled if not necessary to speed up
            process_ms: can be disabled if not necessary to speed up
        """
        super().__init__()
        self.setup_required = True

        if method not in ALLOWED_METHODS:
            raise ValueError(f'Method {method} is not available. Only {ALLOWED_METHODS} are allowed')
        self._method = method
        self._dtype = dtype
        self._process_pan = process_pan
        self._process_ms = process_ms

        self._per_channel = per_channel
        # if ms is scaled per-channel, we cannot scale pan together with all the channels
        self._pan_separate = pan_separate or per_channel
        # Valid only for method = 'meanstd'
        if type(std_width) not in (int, float) or std_width <=0:
            self._std_width=std_width
            raise ValueError('Only positive std_width is allowed, otherwise the values of min and max are undefined')
        self._std_width = std_width
        # Valid only for method = 'histogram'
        if type(hist_cut) not in (int, float) or hist_cut < 0 or hist_cut >= 1:
            self._hist_cut = hist_cut
            raise ValueError('Hist_cut value must be in range [0:1)')
        self._hist_cut = hist_cut

        # ============ total image statistics used for calculations ===========#
        self._pan_min_value = 0
        self._pan_max_value = 1
        self._ms_min_value = [] if self._per_channel else 0
        self._ms_min_value = [] if self._per_channel else 1

        self._ms_mean_value = [] if self._per_channel else 0
        self._ms_std_value = [] if self._per_channel else 0
        self._pan_mean_value = 0
        self._pan_std_value = 0
        # ================ temp storage for image statistics ================ #
        # separate pan mins and maxs for pan and ms
        self._pan_mins = []
        self._pan_maxs = []

        # ms mins and maxs - numeric if not self.per_channel and list else
        self._ms_mins = []
        self._ms_maxs = []

        # mean, std and mean error
        self._pan_means = []
        self._pan_stds = []
        self._pan_mes = []
        self._pan_nums = []

        self._ms_means = []
        self._ms_stds = []
        self._ms_nums = []
        self._ms_mes = []

        # histogram
        self._pan_hist = []
        self._ms_hist = []

    def __repr__(self):
        repr = f'LinearBrightnessScale(method: {self._method}, ' \
               f'pan_separate: {self._pan_separate}, per_channel: {self._per_channel}, ' \
               f'process_pan: {self._process_pan}, process_ms: {self._process_ms})'
        if self._method == 'meanstd':
            repr = repr + f', width: {self._std_width}'
        elif self._method == 'historgam':
            repr = repr + f', {self._hist_cut}'

        return repr

    def setup_from_patch(self, pan, ms):

        if self._method == 'minmax':
            self._minmax_from_patch(pan, ms)
        elif self._method == 'meanstd':
            self._meanstd_from_patch(pan, ms)
        elif self._method == 'historgam':
            self._histogram_from_patch(pan, ms)
        else:
            raise ValueError(f'Unknown method {self._method}')

    def finalize_setup(self):
        """
            Claculates total min and max for the pan and ms values from accumulated windowed min and max
        """
        if self._method == 'minmax':
            self._finalize_minmax()
        elif self._method == 'meanstd':
            self._finalize_meanstd()
        elif self._method == 'historgam':
            self._finalize_histogram()
        else:
            raise ValueError(f'Unknown method {self._method}')

    def process(self, pan, ms):
        """
        Args:
            pan: panchromatic image, 2-dimensional numpy array
            ms: multispectral image, 3-dimensional numpy array, channels-first (rasterio format), as read in worker

        Returns:
            pansharpened image, the same size as pan, but with number of channels as in ms, with IHS method
        """

        dtype = self._dtype if self._dtype is not None else pan.dtype
        pan_res = None
        ms_res = None

        if self._process_pan:
            pan_res = linear_brightness_scale(pan,
                                          self._pan_min_value, self._pan_max_value,
                                          dtype=dtype)
        if self._process_ms:
            if self._per_channel:
                ms_res = np.zeros(shape=ms.shape, dtype=dtype)
                for channel in range(ms.shape[0]):
                    ms_res[channel] = linear_brightness_scale(ms[channel],
                                                          self._ms_min_value[channel], self._ms_max_value[channel],
                                                          dtype=dtype)
            else:
                ms_res = linear_brightness_scale(ms,
                                             self._ms_min_value, self._ms_max_value,
                                             dtype=dtype)
        return pan_res, ms_res

    # =========================== Private functions: methods of setup ======================= #

    # find minimum and maximum value of the whole image
    def _minmax_from_patch(self, pan, ms):

        if not self._pan_separate:
            local_min = min(pan.min(), ms.min())
            local_max = max(pan.max(), ms.max())
            self._pan_mins.append(local_min)
            self._pan_maxs.append(local_max)
            self._ms_mins.append(local_min)
            self._ms_maxs.append(local_max)
        else:
            self._pan_mins.append(pan.min())
            self._pan_maxs.append(pan.max())

            if self._per_channel:
                local_min = []
                local_max = []
                for channel in ms:
                    local_min.append(channel.min())
                    local_max.append(channel.max())
                self._ms_mins.append(local_min)
                self._ms_maxs.append(local_max)
            else:
                self._ms_mins.append(ms.min())
                self._ms_maxs.append(ms.max())

    def _finalize_minmax(self):
        if len(self._pan_mins) == 0 or len(self._pan_maxs) == 0 \
                or len(self._ms_mins) == 0 or len(self._ms_maxs) == 0:
            raise ValueError('Setup was unsuccessful: could not calculate min or max values of the image')
            # TODO: manage the default values?
        self._pan_min_value = np.min(self._pan_mins)
        self._pan_max_value = np.max(self._pan_maxs)
        if self._per_channel:
            self._ms_min_value = np.min(self._ms_mins, axis=0)
            self._ms_max_value = np.max(self._ms_maxs, axis=0)
        else:
            self._ms_min_value = np.min(self._ms_mins)
            self._ms_max_value = np.max(self._ms_maxs)

        # Clear the statistics accumulators
        # separate pan mins and maxs for pan and ms
        self._pan_mins = []
        self._pan_maxs = []

        # ms mins and maxs - numeric if not self.per_channel and list else
        self._ms_mins = []
        self._ms_maxs = []


    # ================== Histogram method ========================= #
    # Find minimum and maximum from histogram
    def _histogram_from_patch(self, pan, ms):
        return

    def _finalize_histogram(self):
        return

    # ======================= Meanstd method ======================== #
    # Find maximum and minimum as mean +- N*std
    @staticmethod
    def _meanstdme(img):
        mean = img.mean()
        std = img.std()
        num = img.size
        me = (img - mean).sum() / num

        return mean, std, num, me

    @staticmethod
    def _totalmeanstd(means, stds, nums, mes):
        mean = np.sum([m*n for m,n in zip(means, nums)])/np.sum(nums)
        # Calculation of the stddev from the window statistics
        std = sqrt(np.sum(
              [(nums[i]*(stds[i]**2 + (mean - means[i])**2 + 2*(means[i] - mean)*mes[i]))
               for i in range(len(means))])/np.sum(nums))
        return mean, std

    def _meanstd_from_patch(self, pan, ms):

        # We also need min and max values if we use meanstd
        self._minmax_from_patch(pan, ms)

        if not self._pan_separate:
            mean, std, num, me = self._meanstdme(np.concatenate([ms, np.expand_dims(pan, 0)], axis=0))
            self._pan_means.append(mean)
            self._pan_stds.append(std)
            self._pan_nums.append(num)
            self._pan_mes.append(me)

            self._ms_means.append(mean)
            self._ms_stds.append(std)
            self._ms_nums.append(num)
            self._ms_mes.append(me)
        else:
            mean, std, num, me = self._meanstdme(pan)
            self._pan_means.append(mean)
            self._pan_stds.append(std)
            self._pan_mes.append(me)
            self._pan_nums.append(num)

            if self._per_channel:
                means = []
                stds = []
                nums = []
                mes = []
                for ch in ms:
                    mean, std, num, me = self._meanstdme(ch)
                    means.append(mean)
                    stds.append(std)
                    nums.append(num)
                    mes.append(me)
                self._ms_means.append(np.array(means))
                self._ms_stds.append(np.array(stds))
                self._ms_mes.append(np.array(mes))
                self._ms_nums.append(np.array(nums))
            else:
                mean, std, num, me = self._meanstdme(ms)
                self._ms_means.append(mean)
                self._ms_stds.append(std)
                self._ms_mes.append(me)
                self._ms_nums.append(num)

    def _finalize_meanstd(self):
        # We need all the statistics to be calculated for every patch in the same order, so we check the length
        patches_num = set(len(stat) for stat in [self._ms_means, self._ms_stds, self._ms_nums, self._ms_mes,
                      self._pan_means, self._pan_stds, self._pan_nums, self._pan_mes])
        if len(patches_num) != 1:
            raise ValueError('Some of the patches staticstics were not calculated correcty, '
                             'and the length of the patch statistics arrays are different')
        elif 0 in patches_num:
            raise ValueError('No data received during the setup')
        # We find the actual min and max of images
        self._finalize_minmax()
        if self._per_channel:
            ms_mean = np.zeros_like(self._ms_means[0])
            ms_std = np.zeros_like(self._ms_stds[0])
            for ch in range(ms_mean.shape[0]):
                ms_mean[ch], ms_std[ch] = self._totalmeanstd(np.array(self._ms_means)[:,ch],
                                                             np.array(self._ms_stds)[:,ch],
                                                             np.array(self._ms_nums)[:,ch],
                                                             np.array(self._ms_mes)[:,ch])

        else:
            ms_mean, ms_std = self._totalmeanstd(self._ms_means, self._ms_stds, self._ms_nums, self._ms_mes)
        pan_mean, pan_std = self._totalmeanstd(self._pan_means, self._pan_stds, self._pan_nums, self._pan_mes)

        self._pan_mean_value = pan_mean
        self._pan_std_value = pan_std

        self._ms_mean_value = ms_mean
        self._ms_std_value = ms_std
        # Then we make the min higher and the max lower if it is necessary
        # Maybe not use minmax, and just clip to the value range?

        axis = 0 if self._per_channel else None
        self._pan_min_value = max(pan_mean - self._std_width * pan_std, self._pan_min_value)
        self._ms_min_value = np.max([ms_mean - self._std_width * ms_std, self._ms_min_value], axis=axis)
        self._pan_max_value = min(pan_mean + self._std_width * pan_std, self._pan_max_value)
        self._ms_max_value = np.min([ms_mean + self._std_width * ms_std, self._ms_max_value], axis=axis)

        # Clear the statistics accumulators
        # separate pan mins and maxs for pan and ms
        self._pan_mins = []
        self._pan_maxs = []

        # ms mins and maxs - numeric if not self.per_channel and list else
        self._ms_mins = []
        self._ms_maxs = []

        # mean, std and mean error
        self._pan_means = []
        self._pan_stds = []
        self._pan_mes = []
        self._pan_nums = []

        self._ms_means = []
        self._ms_stds = []
        self._ms_nums = []
        self._ms_mes = []





import os
import rasterio
import numpy as np
import aeronet.dataset as ds
from numpy.ma import MaskedArray
from typing import List
from aeronet.converters.split import split
from pysharpen.methods import ImgProc


class Worker:
    def __init__(self, methods: List[ImgProc],
                 window_size=(2048, 2048),
                 resampling='bilinear', out_dtype=None):

        self.out_dtype = out_dtype
        self.window_size = window_size
        self.resampling = resampling
        # methods have to be initialized
        self.methods = methods

        # bound is specified by the methods
        self.setup_bound = np.max([m.setup_bound for m in self.methods])
        self.processing_bound = np.max([m.processing_bound for m in self.methods])
        # all the functions of the methods are bound into a single processing function to calculate in a single pass
        def processing_fn(collection):
            if collection.count > 1:
                img = collection.numpy()
                pan = img[-1]
                ms = img[:-1]
            else:
                raise ValueError('The samples inside the collection must be a list '
                                 'with 2 or more channels where PAN channel is the last')
            for m in self.methods:
                pan, ms = m.process(pan, ms)
            return ms
        self.processing_fn = processing_fn

    def setup_methods(self, bc, channels):

        sampler = ds.io.SequentialSampler(bc, channels, sample_size=self.window_size, bound=self.setup_bound)
        # Apply different bound for different calculations to avoid errors
        # At the moment, the bound is maximum of all methods' working bounds
        # Also maybe different bounds for setup and processing

        for sample, block in sampler:
            img = sample.numpy()
            ms = img[:-1]
            pan = img[-1]
            # TODO: apply nodata masks to exclude values
            # if bc.nodata is not None:
            #    pan = MaskedArray(pan, (pan == bc[-1].nodata), fill_value=bc[-1].nodata)
            #    ms = MaskedArray(ms, (ms == bc.nodata), fill_value=bc.nodata)

            for m in self.methods:
                if m.setup_required:
                    # The excessive boundary can affect the statistics of the image, so we cut the input for each method
                    # according to their minimum necessary bound
                    cut = self.setup_bound - m.processing_bound
                    m.setup_from_patch(pan[cut: pan.shape[0]-cut, cut: pan.shape[1]-cut],
                                       ms[:, cut:ms.shape[1]-cut, cut:ms.shape[2]-cut])

        for m in self.methods:
            if m.setup_required:
                m.finalize_setup()

    def process(self, bc, channels, output_labels, output_dir, verbose=False):
        if self.out_dtype is not None:
            dtype = self.out_dtype
        else:
            dtype = bc[-1].dtype
        return ds.Predictor(channels, output_labels, self.processing_fn,
                            sample_size=self.window_size, bound=self.processing_bound,
                            verbose=verbose, dtype=dtype)\
            .process(bc, output_dir)

    def process_separate(self, input_dir,  mul_channels, pan_channel='PAN', output_dir=None, output_labels=None,
                         extensions=('tif', 'tiff', 'TIF', 'TIFF'), verbose=False):
        """
        Processing of the bands represented as a set separate files, one for each band
        Args:
            input_dir:
            mul_channels:
            pan_channel:
            output_dir:
            output_labels:
            extensions:
            verbose:

        Returns:

        """

        # We call pansharpened channel with 'P' prefix, like pansharpen RED is PRED
        # if they are not specified or with errors
        if output_labels is None or len(output_labels) != len(mul_channels):
            if verbose:
                print('Generating names for pansharpened channels')
            output_labels = ['P' + channel for channel in mul_channels]
        # A single band collection is necessary for the Predictor
        # reproject_to provides geographic matching of pan and mul images
        pan_band = ds.Band(ds.parse_directory(input_dir, [pan_channel], extensions)[0])
        mul_bands = ds.BandCollection(ds.parse_directory(input_dir, mul_channels, extensions))

        tmp_names = [os.path.join(input_dir, 'resampled_' + channel + '.tif') for channel in mul_channels]
        tmp_bands = [b.reproject_to(pan_band,
                                    fp=tmp_name,  # We manage the temp files by hand
                                    interpolation=self.resampling) for b, tmp_name in zip(mul_bands, tmp_names)]

        all_bands = ds.BandCollection(tmp_bands + [pan_band])

        self.setup_methods(all_bands, mul_channels + [pan_channel])
        self.process(all_bands, mul_channels + [pan_channel], output_labels, output_dir, verbose)
        # remove temp files
        for band in tmp_bands:
            band._tmp_file = True

    def process_single(self, pan_file, ms_file, out_file, channels=None, clean=True, verbose=False):
        """
        Processing of files where the MS image is in one file and Pan in another.
        It splits the MS file to a set of bands and then as in process_separate.
        The result is then stacked back to a single MS file
        Args:
            pan_file:
            ms_file:
            out_file:
            channels: names for the channels of ms file to split. If None, the names are generated
            clean: if True, all separate band files (both source and pansharpened) are deleted

        Returns:

        """

        folder = os.path.dirname(os.path.abspath(ms_file))
        pan_band = ds.Band(pan_file)

        with rasterio.open(ms_file) as src:
            ms_profile = src.profile
        with rasterio.open(pan_file) as src:
            pan_profile = src.profile

        if channels is None:
            channels = ['B' + str(i+1).zfill(2) for i in range(ms_profile['count'])]
        mul_bands = split(ms_file, folder, channels)

        output_labels = ['P' + channel for channel in channels]
        tmp_names = [os.path.join(folder, 'resampled_' + channel + '.tif') for channel in channels]
        tmp_bands = [b.reproject_to(pan_band,
                                    fp=tmp_name, # We manage the temp files by hand
                                     interpolation=self.resampling) for b, tmp_name in zip(mul_bands, tmp_names)]

        all_bands = ds.BandCollection(tmp_bands + [pan_band])

        self.setup_methods(all_bands, channels + [pan_band.name])
        out_bc = self.process(all_bands, channels + [pan_band.name], output_labels, folder)

        pan_profile.update(count=ms_profile['count'], dtype=out_bc[0].dtype)

        # We want to merge the channels back into one file as it was
        # It requires reading full files, so it's not memory-efficient
        # However, just the read-write operations will not consume more than the image size
        with rasterio.open(out_file, 'w', **pan_profile) as dst:
            for band_num, band in enumerate(out_bc):
                dst.write(band.numpy(), band_num + 1)

        if clean:
            # Mark the files for removal
            for band, pband, tmp_band in zip(mul_bands, out_bc, tmp_bands):
                band._tmp_file = True
                pband._tmp_file = True
                tmp_band._tmp_file = True



import numpy as np
from pysharpen.methods.img_proc import ImgProc
from pysharpen.functional import gamma_correction


class GammaCorrection(ImgProc):
    """
     IHS pansharpening is for RGB image only. It transforms RGB into IHS color space,
     then replaces Intensity with PAN channel and transforms backwards. Based on OpenCV color transforms
    """

    def __init__(self, gamma=2.2, dtype=None,
                 per_channel=False, pan_separate=False,
                 process_pan=True, process_ms=True,
                 **kwargs):
        """

        Args:
            method: minmax, or
            dtype:
            per_channel:
            pan_separate:
            process_pan: can be disabled if not necessary to speed up
            process_ms: can be disabled if not necessary to speed up
        """
        super().__init__()
        self.gamma = gamma
        self.dtype = dtype
        self.process_pan = process_pan
        self.process_ms = process_ms

        self.per_channel = per_channel
        # if ms is scaled per-channel, we cannot scale pan together with all the channels
        self.pan_separate = pan_separate or per_channel

        self._mins = []
        self._maxs = []
        self._min_value = 0
        self._max_value = 255

        self.setup_required = True

    def setup_from_patch(self, pan, ms, nodata = None):
        if pan.dtype == ms.dtype == np.uint8:
            return

        self._mins.append(min(pan.min(), ms.min()))
        self._maxs.append(max(pan.max(), ms.max()))

    def finalize_setup(self):
        """
            Claculates total min and max for the pan and ms values from accumulated windowed min and max
        """
        if len(self._mins) == 0 or len(self._maxs) == 0:
            self._min_value = 0
            self._max_value = 1
            return
        self._max_value = np.max(self._maxs)
        self._min_value = np.max(self._mins)
        self._mins = []
        self._maxs = []

    def process(self, pan, ms):
        if self.process_pan:
            pan = gamma_correction(pan, self.gamma, self._min_value, self._max_value, self.dtype)
        if self.process_ms:
            ms = gamma_correction(ms, self.gamma, self._min_value, self._max_value, self.dtype)
        return pan, ms




import cv2
import numpy as np
from pysharpen.methods.img_proc import ImgProc
from pysharpen.functional import linear_brightness_scale


class IHSPansharpening(ImgProc):
    """
     IHS pansharpening is for RGB image only. It transforms RGB into IHS color space,
     then replaces Intensity with PAN channel and transforms backwards. Based on OpenCV color transforms
    """

    def __init__(self):
        super().__init__()

        self._mins = []
        self._maxs = []
        self._min_value = 0
        self._max_value = 255

        self.setup_required = True

    def setup_from_patch(self, pan, ms, nodata=None):
        if pan.dtype == ms.dtype == np.uint8:
            return
        #self.nodata = nodata
        #if nodata is not None:
        #    pan = MaskedArray(pan, (pan == nodata), fill_value=nodata)
        #    ms = MaskedArray(pan, (ms == nodata), fill_value=nodata)
        self._mins.append(min(pan.min(), ms.min()))
        self._maxs.append(max(pan.max(), ms.max()))

    def finalize_setup(self):
        """
            Claculates total min and max for the pan and ms values from accumulated windowed min and max
        """
        if len(self._mins) == 0 or len(self._maxs) == 0:
            self._min_value = 0
            self._max_value = 1
            return
        self._max_value = np.max(self._maxs)
        self._min_value = np.max(self._mins)
        self._mins = []
        self._maxs = []

    def process(self, pan, ms):

        """

        Args:
            pan: panchromatic image, 2-dimensional numpy array
            ms: multispectral image, 3-dimensional numpy array, channels-first (rasterio format), as read in worker

        Returns:
            pansharpened image, the same size as pan, but with number of channels as in ms, with IHS method
        """

        if pan.shape != ms.shape[1:]:
            raise ValueError('Shapes of multispectral and panchrom are different')
        if ms.shape[0] != 3:
            raise ValueError('IHS pansharpening is restricted to 3 channels, use GIHS for others')
        if ms.dtype != pan.dtype:
            raise ValueError(f'Data types of multispectral {ms.dtype} and panchrom {pan.dtype} are different')

        dtype = pan.dtype
        ms = ms.transpose(1, 2, 0)

        if dtype != np.uint8:
            # Opencv works with uint8 and float32. All other types are converted to float32
            # and translated to 0:1 range.
            # Float32 images are treated the same way in order to manage the value range
            # We assume that the pan and rgb images are in the same brightness range.
            # Histogram equalization can be managed as an additional step, if necessary
            if self._min_value == self._max_value:
                # if one of the images is constant, we do nothing
                return pan, ms
            ms = linear_brightness_scale(ms, self._min_value, self._max_value, dtype=np.float32)
            pan = linear_brightness_scale(pan, self._min_value, self._max_value, dtype=np.float32)
        hsv = cv2.cvtColor(ms, cv2.COLOR_RGB2HSV)
        hue = hsv[:, :, 0]
        saturation = hsv[:, :, 1]
        HSpan = np.array([hue, saturation, pan]).transpose(1, 2, 0)
        RGBpan = cv2.cvtColor(HSpan, cv2.COLOR_HSV2RGB)
        if not (dtype == np.uint8 or dtype == np.float32):
            RGBpan = linear_brightness_scale(RGBpan, 0, 1, dtype=dtype,
                                             out_range=(self._min_value, self._max_value))
        return pan, RGBpan.transpose(2, 0, 1)



import numpy as np
import warnings
from .dtype_utils import value_range, saturate_cast

"""
The purpose of these methods is to scale the image values to enhance the image brightness.

The image type remains unchanged.
If not specified, the range of the output image is 0:1 for floating-point and full allowed image range for integer

We do not assume the input value range because... why, exactly?
"""


def linear_brightness_scale(img, input_min, input_max,
                            dtype=None, out_range=None):
    """
    linearly rescales values of the images from [input_min, input_max] to [out_min, out_max] range
     img:
    :param input_min:
    :param input_max:
    :param dtype
    :param out_range:
    :return:
    """
    if dtype is None:
        dtype = img.dtype
    if out_range is None:
        out_range = value_range(dtype)
    elif value_range(dtype)[0] > out_range[0] \
        or value_range(dtype)[1] < out_range[1]:
        raise ValueError('Out range must fit the out data type')

    if input_max < input_min or out_range[0] > out_range[1]:
        raise ValueError('Maximum value of must not be less than minimum')
    if input_max == input_min or out_range[0] == out_range[1]:
        warnings.warn('Minimum and maximum value of the initial range are equal, scaling is undetermined, '
                      'returning array of minimum value of out range')
        return np.ones_like(img)*img.dtype(out_range[0])

    # Here we preserve dtype - only multiplication
    if np.issubdtype(dtype, np.integer):
        out_img = np.floor_divide(
            np.multiply((img - input_min), (out_range[1] - out_range[0]),
                        dtype=np.float32),
            (input_max - input_min)) + out_range[0]
    else:
        out_img = np.divide(
            np.multiply((img - input_min), (out_range[1] - out_range[0]),
                        dtype=np.float32),
            (input_max - input_min)) + out_range[0]
    return saturate_cast(out_img , dtype)


def gamma_correction(img, gamma, input_min=0, input_max=None, dtype=None):
    """
    Gamma-correction, non-linear brightness shift
    Args:
        img: input image
        gamma: the 1/power of the transform
        input_min: this value remains constant, and all lower values are clipped
        input_max: this value remains constant, and all higher values are clipped
        dtype: output data type. The same as input by default
    Returns:

    """
    if dtype is None:
        dtype = img.dtype
    if input_max is None:
        input_max = img.max()

    # gamma is ill-defined out of the range
    img = img.clip(input_min, input_max)

    return saturate_cast((img / input_max-input_min) ^ (1 / gamma) * input_max + input_min, dtype)





from aeronet.dataset import BandCollectionSample


class ImgProc:
    def __init__(self):
        self.setup_required = False
        self.nodata = None

    def setup_from_patch(self, pan, ms):
        return

    def finalize_setup(self):
        return

    def process(self, pan, ms):
        """

        Args:
            pan: panchromatic image, 2-dimensional numpy array
            ms: multispectral image, 3-dimensional numpy array, channels-first (rasterio format), as read in worker

        Returns:
            pansharpened image, the same size as pan, but with number of channels as in ms
        """

        raise NotImplementedError

    def get_sharpen_fn(self):
        """
        """

        def fn(collection: BandCollectionSample):
            if collection.count > 1:
                img = collection.numpy()
                pan = img[-1]
                ms = img[:-1]
            else:
                raise ValueError('The samples inside the collection must be a list '
                                 'with 2 or more channels where PAN channel is the last')
            return self.process(pan, ms)

        return fn

    @property
    def processing_bound(self):
        """
        required bound for this processing
        :return:
        """
        return 0

    @property
    def setup_bound(self):
        """
        required bound for setup of this function
        :return:
        """
        return 0



import numpy as np
from .brightness_adjustment import linear_brightness_scale
from .dtype_utils import value_range
"""
The purpose of this module is to change the data types of the images in convenient way.
These are not the default numpy-like methods of casting but the ways to convert image type without loss of quality
or even with 

Also includes utils to handle numpy dtypes better
"""


def linear_cast_meanstd(img, dtype, width=3.0,
                        out_min=None, out_max=None):
    mean = np.mean(img)
    std = np.std(img)

    in_min = max(0., mean - width * std)
    in_max = min(img.max(), mean + width * std)
    if out_min is not None and out_max is not None:
        out_range = (out_min, out_max)
    else:
        out_range = value_range(dtype)
    return linear_brightness_scale(img, in_min, in_max, dtype, out_range)


def linear_cast_clip(img, dtype, percent=0.1,
                     out_min=None, out_max=None):

    channels = min(img.shape)
    data = img.flatten()
    V = data.shape[0] / channels
    dv = V * percent / 100

    hist = np.histogram(data, bins=data.max() - data.min())[0]

    tail = 0
    in_min = 0
    while tail < dv:
        tail += hist[in_min]
        in_min += 1

    tail = 0
    in_max = data.max() - data.min() - 1
    while tail < dv:
        tail += hist[in_max]
        in_max -= 1

    if out_min is not None and out_max is not None:
        out_range = (out_min, out_max)
    else:
        out_range = value_range(dtype)
    return linear_brightness_scale(img, in_min, in_max, dtype, out_range)



import numpy as np
from pysharpen.methods.img_proc import ImgProc
from pysharpen.functional import saturate_cast


class BroveyPansharpening(ImgProc):
    """
    Brovey transform for pansharpening.
    Formula for the output is: MSpan[i] = MS[i] * pan / mean (MS)
    Mean is weighted, default weights are equal.
    """

    def __init__(self, weights=None):
        """
        :param weights:
        """
        super().__init__()
        self.weights = None

        if weights is not None:
            self.count = len(weights)
            weights = np.array([[weights]]).transpose(2, 1, 0)
            self.weights = weights / sum(weights)
        else:
            self.weights = 1
        self.setup_required = False

    def _calculate_ratio(self, pan, ms):
        return pan / (np.sum(ms * self.weights, 0))

    def process(self, pan, ms):
        ratio = self._calculate_ratio(pan, ms)
        res = ratio * ms
        return pan, saturate_cast(res, ms.dtype)



import numpy as np
from pysharpen.methods.img_proc import ImgProc
from pysharpen.functional import saturate_cast, wider_type


class GIHSPansharpening(ImgProc):
    """
    Generalized IHS-like transform for any number of channels.
    Formula for the output is: MSpan[i] = MS[i] + pan - mean (MS)
    Similar to Brovey, but additive rather than multiplicative
    """

    def __init__(self):
        super().__init__()

    def process(self, pan, ms):
        """
        :param pan:
        :param ms:
        :return:
        """
        # The methods expect already resized images
        if pan.shape != ms.shape[1:]:
            raise ValueError('Shapes of multispectral and panchrom are different')
        ms = ms.transpose(1, 2, 0)

        ms_pansharpened = ms.astype(wider_type(pan.dtype))
        mean = np.expand_dims(np.sum(ms_pansharpened, axis=2) / ms_pansharpened.shape[0], 2)
        ms_pansharpened = ms_pansharpened + np.expand_dims(pan.astype(wider_type(ms.dtype)), 2) - mean

        ms_pansharpened = saturate_cast(ms_pansharpened, ms.dtype)
        return pan, ms_pansharpened.transpose(2, 0, 1)


import numpy as np
"""
The purpose of this module is to change the data types of the images in convenient way.
These are not the default numpy-like methods of casting but the ways to convert image type without loss of quality
or even with 

Also includes utils to handle numpy dtypes better
"""


def value_range(dtype):
    """
    returns a tuple (min value, max value) for integer dtype
    or (0,1) for float dtype
    :param dtype: numpy data type or its string representation
    :return:
    """
    if np.issubdtype(dtype, np.integer):
        out_range = (np.iinfo(dtype).min, np.iinfo(dtype).max)
    else:
        out_range = (0., 1.)
    return out_range


def wider_type(dtype):
    """
    Returns the dtype which can handle arithmetic with the specified dtype without overflow.
    However, for the multiplication/division it is more convenient to cast to float32 in any case

    # TODO: decide whether we need or np.result_type is enough
    :param dtype: numpy data type or its string representation
    :return:
    """
    if dtype == np.int8 or dtype == np.uint8:
        return np.int16
    if dtype == np.int16 or dtype == np.uint16:
        return np.int32
    else:
        # actually, we may assume that other types are unlikely and wide enough to fit all the calculations
        return dtype


def saturate_cast(img, dtype):
    """
    Implementation of opencv saturate_cast, which changes datatype and clips the data to the output type range
    :param img: input image
    :param dtype: data type of output image
    :return: new image of dtype
    """
    if np.issubdtype(dtype, np.integer):
        return np.clip(np.around(img,0), np.iinfo(dtype).min, np.iinfo(dtype).max).astype(dtype)
    else:
        return img.astype(dtype)
